/* MODUL VF */
/* Verschiedene Verteilungsfunktionen */

#include <math.h>
#include "vfm.h"

double chi2(long n, double x)
/* Verteilungsfunktion der chi2-Verteilung mit n Freiheitsgraden */
/* Benoetigt Funktion Phi */
/* Benoetigt M_PI */
{
double f, c, p;
long k;

f=1; c=-exp(-x/2); k=n%2;
if (k!=0)
{
  p=Phi(-sqrt(x)); f=f-2*p;
  c*=sqrt(2*x/M_PI);
}
do
{
  k+=2;
  if (k<=n)
  {
    f=f+c; c=c*x/k;
  }
}
while (k<=n);
if (f>1e-100)
  return (f);
else
  return (1e-100);
}

double chiinv (long n, double y)
/* Inverse der chi2-Verteilung mit n Freiheitsgraden */
/* Benoetigt Funktion chi2 */
{
double a=0, b=2, c;

  do
  {
    b*=2;
  }
  while (chi2(n,b)<y);
  do
  {
    c=(a+b)/2;
    if (chi2(n,c)<y)
      a=c;
    else
      b=c;
  }
  while (fabs(a-b)>1e-5);
  return(c);
}

double Phi (double x)
/* Verteilungsfunktion der Normalverteilung */
/* Benoetigt M_PI */
{
const double a1=0.31938153,
	     a2=-0.356563782,
	     a3=1.781477937,
	     a4=-1.821255978,
	     a5=1.330274429,
	     c=0.2316419;

double t, f, ph;

t=1/(1+c*fabs(x));
f=t*(a1+t*(a2+t*(a3+t*(a4+t*a5))));
ph=exp(-x*x/2)/sqrt(2*M_PI);
ph*=f;
if (x>0)
  return (1-ph);
else
  return (ph);
}

double phiinv(double al)
/* Inverses der VF der Normalverteilung */
{
const double a0=0.322232431088,
	     a1=1.0,
	     a2=0.342242088547,
	     a3=0.0204231210245,
	     a4=0.453642210148e-4,
	     b0=0.099348462606,
	     b1=0.588581570495,
	     b2=0.531103462366,
	     b3=0.10353775285,
	     b4=0.0038560700634;

double t,z,b;

  b=al;
  if (al>0.5)
    b=1-al;
  if (b<1e-9)
    b=1e-9;
  t=sqrt(-2*log(b));
  z=-t+(a0+a1*t+a2*t*t+a3*t*t*t+a4*t*t*t*t)/(b0+b1*t+b2*t*t+b3*t*t*t+b4*t*t*t*t);
  if (al<0.5)
    return(z);
  else
    return (-z);
}

double ffr (long m, long n, double gam)
/* Fraktile der F-Verteilung */
/* Benoetigt Funktionen fv, lng */
{
  double a, b, c;

  a=1e-10; b=1e+10;
  do
  {
    c=(a+b)/2;
    if (fv(m, n, c)>gam)
      b=c;
    else
      a=c;
  }
  while (fabs(b-a)>1e-9);
  return (c);
}

double fv (long m, long n, double a)
/* Verteilungsfunktion F-Verteilung */
/* Benoetigt Funktion lng */
{
long r, k, con;
double x, p, q, x1, y1, p1, q1, c, s, e, b, g1, g2, g3, ld, d, i;

  x=m*a/(m*a+n); p= (double) m/2.0; q= (double) n/2.0;
  x1=x; y1=1.0-x; p1=p; q1=q;
  if (p<(x*(p+q)))
  {
    x1=1.0-x; y1=x; p1=q; q1=p;
  }
  k=1; c=1.0; s=1.0; e=x1/y1; b=q1-1.0; r=1.0+q1+y1*(p1+q1);
  if (k==r)
    e=x1;
  do
  {
    c=c*e*b/(p1+k); con=0;
    if (c>1e-12)
    {
      con=1; s+=c;
      ++k;
      if (k<r+1)
      {
	b-=1.0;
	if (k==r)
	  e=x1;
      }
      else
      {
	if (k==r+1)
	  b=p1+q1;
	else
	  b+=1.0;
      }
    }
  }
  while (con!=0);
  g1=lng(p1+q1); g2=lng(p1); g3=lng(q1);
  ld=p1*log(x1)+q1*log(y1)+g1-g2-g3;
  d=exp(ld); i=d*s/(p1*y1);
  if (p<(x*(p+q)))
    i=1.0-i;
  return (i);
}

double tw (long n, double x)
/* Verteilungsfunktion t-Verteilung */
/* Benoetigt M_PI */
{
  long k;
  double a, b, c, f;

  a=x/sqrt(n); b=1+a*a;
  c=a/(2*sqrt(b)); f=0.5;
  k=n%2;
  if (k==1)
  {
    c=a/(b*M_PI); f+=atan(a)/M_PI;
  }
  k+=2;
  while (k<=n)
  {
    f+=c; c*=(1.0-1.0/k)/b; k+=2;
  }
  return (f);
}

double lng(double x)
/* Hilfsfunktion fuer F-Verteilung */
{
const double a0=0.0833333331,
	     a1=-0.00277765546,
	     a2=0.00077783067,
	     c=2.5066283;

double a, z, f;

  z=x; a=1.0;
  while (z<8)
  {
    a/=z; z+=1.0;
  }
  f=(a0+(a1+a2/(z*z))/(z*z))/z;
  return((z-0.5)*log(z)-z+f+log(a*c));
}

double choose(long n, long k)
/* n ueber k */
{
long i, k2=k;
double prod1=1, prod2=1;

  if (n-k<k)  
    k2=n-k;
  for (i=1; i<=k2; ++i){
    prod1*=(n-k2+i);
    prod2*=i;
  }
  /*  printf("n= %d k= %d prod1= %12.3lf  prod2= %12.3lf  choose= %12.3lf\n",n,k,prod1,prod2,prod1/prod2); */
  return(prod1/prod2);
}  

double dbinom(long n, double p, long k)
     /* Binomial(n,p)-density */
     /* Benoetigt Funktion choose */
{
  return(choose(n,k)*pow(p,k)*pow(1-p,n-k));
}

double pbinom(long n, double p, long k)
     /* Binomial(n,p) cdf */
     /* Benoetigt Funktionen dbinom, choose */
{
long i; double sum=0;

  for (i=0; i<=k; ++i)
    sum+=dbinom(n,p,i);
  return(sum);
}

long qbinomsi(long n, double p, double alpha, long imin)
     /* 1, if quantile binom(n,p,alpha)< imin; sonst 0 */
     /* Benoetigt Funktionen pbinom, dbinom, choose */
{
  return(pbinom(n,p,imin-1) < alpha);
}


